/*
Text Receiver
for ESP32 M5Stack Base Module
Elektor / Jens Nickel
My Journey in the IoT
*/





#include <Preferences.h>   // this library is used to get access to Non-volatile storage (NVS) of ESP32
// see https://github.com/espressif/arduino-esp32/blob/master/libraries/Preferences/examples/StartCounter/StartCounter.ino



// +++++++++++++++++++ Start of Display Library +++++++++++++++++++


#define DISPLAY_X_MAX           320
#define DISPLAY_Y_MAX           240
#define DISPLAY_BUFFER_MAX      4
#define DISPLAY_FONTSIZE_MAX    6
#define DISPLAY_CHARSIZE_MAX    512


// Hardware independent

// Display Buffers

int    DisplayBufferCount;          // number of display buffers used (0..4)
String DisplayBufferText[4];        // Text to be displayed  
int    DisplayBufferX[4];           // X position in pixels
int    DisplayBufferY[4];           // Y position in pixels
int    DisplayBufferFontSize[4];    // FontSize 1..5
int    DisplayBufferCharSize[4];    // length of the buffer text in chars (for deleting the text on the display with spaces)



// Initialize the values
void DisplayBufferInit() 
{
  for (int c = 0; c < 4; c++)
  {
    DisplayBufferText[c] = "";
    DisplayBufferX[c] = 0;
    DisplayBufferY[c] = 0;
    DisplayBufferFontSize[c] = 1;
    DisplayBufferCharSize[c] = 0;
  }
  DisplayBufferCount = 0;
}


// set one display buffer with values
// return 1 in case of valid parameters
// return -1 in case of invalid parameters
int DisplayBufferSet(unsigned int DisplayBufferNumber, unsigned int _DisplayBufferX, unsigned int _DisplayBufferY, unsigned int _DisplayBufferFontSize, unsigned int _DisplayBufferCharSize) 
{
  int DisplayBufferError = 1;
  
  if (DisplayBufferNumber >= DISPLAY_BUFFER_MAX)
  {
    DisplayBufferError = -1;
  }

  if (_DisplayBufferX >= DISPLAY_X_MAX)
  {
    DisplayBufferError = -1;
  }

  if (_DisplayBufferY >= DISPLAY_Y_MAX)
  {
    DisplayBufferError = -1;
  }

  if (_DisplayBufferFontSize > DISPLAY_FONTSIZE_MAX)
  {
    DisplayBufferError = -1;
  }

  if (_DisplayBufferCharSize > DISPLAY_CHARSIZE_MAX)
  {
    DisplayBufferError = -1;
  }

  if (DisplayBufferError == 1)
  {
    DisplayBufferText[DisplayBufferNumber] = "";
    DisplayBufferX[DisplayBufferNumber] = _DisplayBufferX;
    DisplayBufferY[DisplayBufferNumber] = _DisplayBufferY;
    DisplayBufferFontSize[DisplayBufferNumber] = _DisplayBufferFontSize;
    DisplayBufferCharSize[DisplayBufferNumber] = _DisplayBufferCharSize;
    if (DisplayBufferCount <= DisplayBufferNumber)
    {
      DisplayBufferCount = DisplayBufferNumber + 1;      // adapt the number of display buffers 
    }
  }

  return DisplayBufferError;
}



// put text in a display buffer
int DisplayBufferPutTextInBuffer(int DisplayBufferNumber, String DisplayText) 
{
  int DisplayBufferError = 1;

  if (DisplayBufferNumber >= DISPLAY_BUFFER_MAX)
  {
    DisplayBufferError = -1;
  }
  
  if (DisplayBufferError == 1)
  {
    unsigned int lg = DisplayText.length();                  // length of text to be displayed in chars
  
    DisplayBufferText[DisplayBufferNumber] = DisplayText;    // put text in the beginning of buffer
         
    for (int i = lg; i < DisplayBufferCharSize[DisplayBufferNumber]; i++)      // fill the rest with spaces
    {
          DisplayBufferText[DisplayBufferNumber] += ' '; 
    }
  }

  return DisplayBufferError;
}



// hardware dependant

#include <M5Stack.h>  // M5Stack Hardware Library (Display, Buttons)

// Setup the display
void DisplaySetup()
{
  M5.begin();            // access to M5 hardware functions from now on

  DisplayBufferInit();   // initialize Display buffers
}



// show the content of display buffer on the display
int DisplayBufferShowTextOfBuffer(int DisplayBufferNumber) 
{  
   int DisplayBufferError = 1;

   if (DisplayBufferNumber >= DISPLAY_BUFFER_MAX)
   {
        DisplayBufferError = -1;
   }

   if (DisplayBufferError == 1)
   {
     M5.Lcd.setCursor(DisplayBufferX[DisplayBufferNumber] , DisplayBufferY[DisplayBufferNumber]);   // set cursor on position of display buffer to show the text
   
     M5.Lcd.setTextSize(DisplayBufferFontSize[DisplayBufferNumber]);
     M5.Lcd.print(DisplayBufferText[DisplayBufferNumber]);                                          // print the text on Display including the spaces to delete the old text
   } 

   return DisplayBufferError;
}



// put text in a display buffer
// and show the content of display buffer on the display
int DisplayBufferPrint(int DisplayBufferNumber, String DisplayText) 
{
  int DisplayBufferError = DisplayBufferPutTextInBuffer(DisplayBufferNumber, DisplayText);     // put text in a display buffer
  
  if (DisplayBufferError == 1)
  {
     DisplayBufferShowTextOfBuffer(DisplayBufferNumber);                                       // show the content of display buffer on the display
  } 
   
  return DisplayBufferError;
}


// +++++++++++++++++++ End of Display library +++++++++++++++++++++



// ++++++++++++++++++ Start of Button library +++++++++++++++++++++

#define BUTTON_MAX      3


// left / middle / right button pressed?
// we can poll only 1 button at the same time
unsigned int ButtonIsPressed(unsigned int ButtonNumber)       
{
  unsigned int ButtonPressed = 0;

  if (M5.BtnA.isPressed() && ButtonNumber == 0)       // left button
  {
    ButtonPressed = 1;
  }
  if (M5.BtnB.isPressed() && ButtonNumber == 1)       // middle button
  {
    ButtonPressed = 1;
  }
  if (M5.BtnC.isPressed() && ButtonNumber == 2)       // right button
  {
    ButtonPressed = 1;
  }

  return ButtonPressed;
}


void ButtonUpdate()                      // this must be called regularly to poll status of buttons
{
  M5.update();
}


// +++++++++++++++++++ End of Button library ++++++++++++++++++++++





// +++++++++++++++++++ Start of Application Code +++++++++++++++++++




// include Param library
#include <Param.h>

// construct the object of class Param for text messages
Param Messages = Param();

// construct the object of class Param for network config parameters
Param NetworkConfig = Param();


// include WiFiWebserver library
#include <WiFiWebserver.h>

// construct the object of class WiFiWebserver
WiFiWebserver myWebserver = WiFiWebserver();


// include WiFi Network library
#include <WiFiNetwork.h>

// construct the object of class WiFiNetwork
WiFiNetwork myWiFiNetwork = WiFiNetwork();



Preferences preferences;   // we must generate this object of the preference library to read and write values in non volatile storage


// variable to hold status of router network login
byte RouterNetworkDeviceState; 
#define NETWORKSTATE_LOGINREQUESTED       2
#define NETWORKSTATE_LOGGED               1
#define NETWORKSTATE_NOTLOGGED            0


// ESP32 in station or access point mode?
byte WifiDeviceState;
#define WIFIDEVICESTATE_STATIONMODE       1
#define WIFIDEVICESTATE_APMODE            2
#define WIFIDEVICESTATE_BOTHMODES         3


// for timing of next attempt to login in router network
byte CounterForLoginRouterNetworkAttempt;


// WebserverURL to be shown on display
String WebserverURL;




// +++++++++++++++++++ Start of GUI functions +++++++++++++++++++++



// Setup of the GUI
// call this function on startup
void GUI_Setup()
{
  DisplaySetup();

  DisplayBufferSet(0, 0,  0, 2, 2);       // displaybuffer for heartbeat character (0/1) to show that ESP32 is working and not blocked with tasks
  DisplayBufferSet(1, 30, 0, 2, 30);      // displaybuffer for webserver URL and network status
  DisplayBufferSet(2, 0, 40, 4, 30);      // displaybuffer for text message (bigger fontsize)

  DisplayBufferPrint(1, "ESP32 Init");    // initial text on display for startup

}


byte HeartbeatCount = 0;


// call this function to show a heartbeat sign 
void GUI_Heartbeat()
{
    HeartbeatCount++;  // 0..255 again and again
    
    DisplayBufferPrint(0, String(HeartbeatCount & 1));   // show only last bit = 0 or 1 on position of displaybuffer 0
}



// call this function to show updated status of the device
void GUI_Update()
{
   String DeviceText = WebserverURL;
   
   if (RouterNetworkDeviceState == NETWORKSTATE_LOGGED)    // logged in network               
   {
      DeviceText += " NETWORK";          
   }
   
   if (RouterNetworkDeviceState == NETWORKSTATE_NOTLOGGED)     // not logged in network     
   {
      DeviceText += " NO NETWORK";     
   }

   if (RouterNetworkDeviceState == NETWORKSTATE_LOGINREQUESTED)   // login requested, if user submits new config values
   {
      DeviceText += " NETW REQ";     
   }

 
   DisplayBufferPrint(1, DeviceText);    // show webserver URL and the status of the device on position of displaybuffer 1

   DisplayBufferPrint(2, Messages.PValue[0]);    // show received text on position of displaybuffer 2

   
}

// +++++++++++++++++++ End of GUI functions +++++++++++++++++++++






void setup() 
{
  
 
  GUI_Setup();

  // Open serial communications and wait for port to open
  Serial.begin(19200);
  Serial.println("Hello ESP32!");

  // put the ESP32 in station mode
  myWiFiNetwork.SetModeNetworkStation();

  WifiDeviceState = WIFIDEVICESTATE_STATIONMODE;

  LoginRouterNetwork();                     // login attempt in router network
  CounterForLoginRouterNetworkAttempt = 0;

  WebserverURL = myWiFiNetwork.GetOwnIPAddressInRouterNetwork();   // get IP address (will be 0.0.0.0 if not logged in)

  // start the webserver to listen for request of clients (in router WLAN or own ESP32 network)
  myWebserver.Start();



  // network config values
  NetworkConfig.Init();
  
  NetworkConfig.PName[0] = "SSID";
  NetworkConfig.PType[0] = 1;    // type of value is text
  
  NetworkConfig.PName[1] = "PASSWORD";
  NetworkConfig.PType[1] = 1;    // type of value is text

  NetworkConfig.PCount = 2;     // we have 2 values

  // take out the 2 values from out the non-volatile storage
  preferences.begin("MyJourney", false);   // see https://github.com/espressif/arduino-esp32/blob/master/libraries/Preferences/examples/StartCounter/StartCounter.ino 

  
  String strSSID          = preferences.getString("SSID", "");
  String strPassword      = preferences.getString("Password", "");
  
  
  // put the NVS stored values in RAM for the program
  NetworkConfig.PValue[0] = strSSID;
  NetworkConfig.PValue[1] = strPassword;
  
  // initialize messages parameter set
  Messages.Init();

  // we have only 1 value - the message to be shown on display
  Messages.PValue[0] = "";
  Messages.PName[0]  = "Text";
  Messages.PType[0]  = 1;     // type of value is text
  Messages.PCount    = 1;       // only 1 value


  GUI_Update();
}



// store SSID and Passwort in the non volatile storage
void ProcessAndValidateConfigValues()
{  
    preferences.putString("SSID",           NetworkConfig.PValue[0] );
    preferences.putString("Password",       NetworkConfig.PValue[1] );
}



// Read incoming request from client (if there is any)
// process values in GET parameter
// and send out new HTML page to client
// returns 1 if GET parameter with new values was received, after URL address 192.168.4.1 (as a sign that user wants to configure device once new)
// returns 0 if any other case 
int GetConfigRequestAndSendHTMLPage()
{
    int ConfigRequestReceived = 0;
    
    String GETParameter = myWebserver.GetRequestGETParameter();   // look for client request
  
    if (GETParameter.length() > 0)        // we got a request, client connection stays open
    {
           
      if (GETParameter.length() > 1)      // request contains some GET parameter
      {
                  
          if (myWebserver.WebRequestHostAddress == "192.168.4.1")                               // the user entered address of ESP32 in own network in a browser, with GET parameter values for configuration
          {
    
              NetworkConfig = myWebserver.DeSerializeFromGETParameter(GETParameter, NetworkConfig);     // decode the GET parameter to get Network Config Values 0 and 1 (SSID and Password)

              ProcessAndValidateConfigValues();                                  // store SSID, Password in non-volatile storage
              
              ConfigRequestReceived = 1;
              
          }

          
          if (myWebserver.WebRequestHostAddress == myWiFiNetwork.GetOwnIPAddressInRouterNetwork() )    // the user entered the address of ESP32 in router network in a browser, with GET parameter value for the message to be shown on display
          {             
              Messages = myWebserver.DeSerializeFromGETParameter(GETParameter, Messages);             // decode the GET parameter to get message to be shown on display, will be stored in Message.PValue[0]
          }
          
      }


      String HTMLPage;

             
      if (myWebserver.WebRequestHostAddress == "192.168.4.1")                  //   the user entered entered address of ESP32 in own network in the browser, to get the configuration webpage               
      {
          // build an HTML Form with a table, and a submit button
          // table contains all the configuration values in text controls (SSID and password)
          HTMLPage = myWebserver.FormHTML("Network",        myWebserver.SerializeToHTMLInputTable(NetworkConfig) )    + "<br>IP Address: " + myWiFiNetwork.GetOwnIPAddressInRouterNetwork()  ;
      }
      
      if (myWebserver.WebRequestHostAddress == myWiFiNetwork.GetOwnIPAddressInRouterNetwork() )   // the user entered the address of ESP32 in router network in a browser
      {
         // build an HTML Form with a table, and a submit button
         // table contains the message in a text control
         HTMLPage = myWebserver.FormHTML("Booth Message",   myWebserver.SerializeToHTMLInputTable(Messages) ); 
      }
      
      // send out the webpage to client = webbrowser and close client connection
      myWebserver.SendHTMLPage(HTMLPage); 
      
  }

  return ConfigRequestReceived;
}






// try to login to router network
void LoginRouterNetwork()
{
  String strSSID = preferences.getString("SSID", "");                 // takeout SSID and Password out of non-volatile storage  
  String strPassword = preferences.getString("Password", "");              

  char* txtSSID = const_cast<char*>(strSSID.c_str());                 // convert to char*
  char* txtPassword = const_cast<char*>(strPassword.c_str());         // https://coderwall.com/p/zfmwsg/arduino-string-to-char

  int successDisconnect = myWiFiNetwork.RouterNetworkDisconnect();             // disconnect from router network

  int successConnect = myWiFiNetwork.RouterNetworkConnect(txtSSID, txtPassword);   // then try to connect once new with new login-data

  // print out local address of ESP32 in Router network (LAN)
  Serial.println(myWiFiNetwork.GetOwnIPAddressInRouterNetwork());
    
  if (successConnect == 1)
  {      
      RouterNetworkDeviceState = NETWORKSTATE_LOGGED;                 // set the state      
  }
  else
  {
      RouterNetworkDeviceState = NETWORKSTATE_NOTLOGGED;              // set the state             
  }   

    
}






void loop() {

   // check if request from client was received (PC or smartphone browser with HTML config form)
   int ConfigRequestReceived = GetConfigRequestAndSendHTMLPage();
   
   if (ConfigRequestReceived == 1)  // yes
   {
      RouterNetworkDeviceState = NETWORKSTATE_LOGINREQUESTED;    // set state variable that a new login attempt follows
      GUI_Update();                             // show this on display
      LoginRouterNetwork();                     // login attempt
      CounterForLoginRouterNetworkAttempt = 0;
   }

  
   
   if (RouterNetworkDeviceState != NETWORKSTATE_LOGGED)    // if we are not logged in or login attempt requested by user
   {
      if (CounterForLoginRouterNetworkAttempt > 40)        // counter for a delay between login attempts (approx. every 2 s)
      {
        
        GUI_Update();                             // update GUI to show login attempt
        
        LoginRouterNetwork();                     // login attempt
        CounterForLoginRouterNetworkAttempt = 0;
      }
      CounterForLoginRouterNetworkAttempt++;
   }

   
  // refresh the URL of ESP32
  if (WifiDeviceState == WIFIDEVICESTATE_BOTHMODES)   // we are in Access Point Mode and Station mode also
  {
    WebserverURL = myWiFiNetwork.GetIPAddressInOwnNetwork();     // get address of ESP32 in own network
  }
  else                                               // we are just in Station mode
  {
    WebserverURL = myWiFiNetwork.GetOwnIPAddressInRouterNetwork();   // get address of ESP32 in router network  
  }


  ButtonUpdate();   // we have to call this regularly to poll status of buttons
  
  if (ButtonIsPressed(0) == 1 && ButtonIsPressed(1) == 1)    // left and middle button pressed at the same time
  {
    if (WifiDeviceState == WIFIDEVICESTATE_BOTHMODES)        // we are in Access Point Mode and Station mode also
    {
      WifiDeviceState = WIFIDEVICESTATE_STATIONMODE;      // so we now get just in station mode
      myWiFiNetwork.SetModeNetworkStation();               
      LoginRouterNetwork();                               // login in router network
      WebserverURL = myWiFiNetwork.GetOwnIPAddressInRouterNetwork();    // get the address in router network
      
    }
    else                                                   // we are just in station mode
    {
      WifiDeviceState = WIFIDEVICESTATE_BOTHMODES;            // so we now get in Access point mode and Station mode also
      
      // put the ESP32 in both modes
      myWiFiNetwork.SetBothModesNetworkStationAndAccessPoint();
    
      // Start access point with SSID "ESP32_MyJourneyIoT"
      int success = myWiFiNetwork.AccessPointStart("ESP32_MyJourneyIoT");
      
      WebserverURL = myWiFiNetwork.GetIPAddressInOwnNetwork();   // get address of ESP32 in own network
      
    }
    GUI_Update();    // show device state and received message on display
    delay(1000);     // wait a second
  }
  
  GUI_Heartbeat();  // show heartbeat sign on display
  GUI_Update();     // show device state and received message on display
  
  delay(50);          // wait 50 ms
}
